// Copyright 2020 The Chromium Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

#import <Firebase/Firebase.h>
#import <TargetConditionals.h>
#import <firebase_core/FLTFirebasePluginRegistry.h>

#import "Private/FLTAuthStateChannelStreamHandler.h"
#import "Private/FLTIdTokenChannelStreamHandler.h"
#import "Private/FLTPhoneNumberVerificationStreamHandler.h"
#import "Private/PigeonParser.h"

#import "Public/CustomPigeonHeader.h"
#import "Public/FLTFirebaseAuthPlugin.h"
@import CommonCrypto;
#import <AuthenticationServices/AuthenticationServices.h>

NSString *const kFLTFirebaseAuthChannelName = @"plugins.flutter.io/firebase_auth";

// Argument Keys
NSString *const kAppName = @"appName";

// Provider type keys.
NSString *const kSignInMethodPassword = @"password";
NSString *const kSignInMethodEmailLink = @"emailLink";
NSString *const kSignInMethodFacebook = @"facebook.com";
NSString *const kSignInMethodGoogle = @"google.com";
NSString *const kSignInMethodTwitter = @"twitter.com";
NSString *const kSignInMethodGithub = @"github.com";
NSString *const kSignInMethodApple = @"apple.com";
NSString *const kSignInMethodPhone = @"phone";
NSString *const kSignInMethodOAuth = @"oauth";

// Credential argument keys.
NSString *const kArgumentCredential = @"credential";
NSString *const kArgumentProviderId = @"providerId";
NSString *const kArgumentProviderScope = @"scopes";
NSString *const kArgumentProviderCustomParameters = @"customParameters";
NSString *const kArgumentSignInMethod = @"signInMethod";
NSString *const kArgumentSecret = @"secret";
NSString *const kArgumentIdToken = @"idToken";
NSString *const kArgumentAccessToken = @"accessToken";
NSString *const kArgumentRawNonce = @"rawNonce";
NSString *const kArgumentEmail = @"email";
NSString *const kArgumentCode = @"code";
NSString *const kArgumentNewEmail = @"newEmail";
NSString *const kArgumentEmailLink = kSignInMethodEmailLink;
NSString *const kArgumentToken = @"token";
NSString *const kArgumentVerificationId = @"verificationId";
NSString *const kArgumentSmsCode = @"smsCode";
NSString *const kArgumentActionCodeSettings = @"actionCodeSettings";

// MultiFactor
NSString *const kArgumentMultiFactorHints = @"multiFactorHints";
NSString *const kArgumentMultiFactorSessionId = @"multiFactorSessionId";
NSString *const kArgumentMultiFactorResolverId = @"multiFactorResolverId";
NSString *const kArgumentMultiFactorInfo = @"multiFactorInfo";

// Manual error codes & messages.
NSString *const kErrCodeNoCurrentUser = @"no-current-user";
NSString *const kErrMsgNoCurrentUser = @"No user currently signed in.";
NSString *const kErrCodeInvalidCredential = @"invalid-credential";
NSString *const kErrMsgInvalidCredential =
    @"The supplied auth credential is malformed, has expired or is not "
    @"currently supported.";

@interface FLTFirebaseAuthPlugin ()
@property(nonatomic, retain) NSObject<FlutterBinaryMessenger> *messenger;
@property(strong, nonatomic) FIROAuthProvider *authProvider;
// Used to keep the user who wants to link with Apple Sign In
@property(strong, nonatomic) FIRUser *linkWithAppleUser;
@property(strong, nonatomic) FIRAuth *signInWithAppleAuth;
@property BOOL isReauthenticatingWithApple;
@property(strong, nonatomic) NSString *currentNonce;
@property(strong, nonatomic) void (^appleCompletion)
    (PigeonUserCredential *_Nullable, FlutterError *_Nullable);
@property(strong, nonatomic) PigeonFirebaseApp *appleArguments;

@end

@implementation FLTFirebaseAuthPlugin {
  // Used for caching credentials between Method Channel method calls.
  NSMutableDictionary<NSNumber *, FIRAuthCredential *> *_credentials;

#if TARGET_OS_IPHONE
  // Map an id to a MultiFactorSession object.
  NSMutableDictionary<NSString *, FIRMultiFactorSession *> *_multiFactorSessionMap;

  // Map an id to a MultiFactorResolver object.
  NSMutableDictionary<NSString *, FIRMultiFactorResolver *> *_multiFactorResolverMap;
#endif

  NSObject<FlutterBinaryMessenger> *_binaryMessenger;
  NSMutableDictionary<NSString *, FlutterEventChannel *> *_eventChannels;
  NSMutableDictionary<NSString *, NSObject<FlutterStreamHandler> *> *_streamHandlers;
  NSData *_apnsToken;
}

#pragma mark - FlutterPlugin

- (instancetype)init:(NSObject<FlutterBinaryMessenger> *)messenger {
  self = [super init];
  if (self) {
    [[FLTFirebasePluginRegistry sharedInstance] registerFirebasePlugin:self];
    _credentials = [NSMutableDictionary<NSNumber *, FIRAuthCredential *> dictionary];
    _binaryMessenger = messenger;
    _eventChannels = [NSMutableDictionary dictionary];
    _streamHandlers = [NSMutableDictionary dictionary];

#if TARGET_OS_IPHONE
    _multiFactorSessionMap = [NSMutableDictionary dictionary];
    _multiFactorResolverMap = [NSMutableDictionary dictionary];
#endif
  }
  return self;
}

+ (void)registerWithRegistrar:(NSObject<FlutterPluginRegistrar> *)registrar {
  FlutterMethodChannel *channel =
      [FlutterMethodChannel methodChannelWithName:kFLTFirebaseAuthChannelName
                                  binaryMessenger:[registrar messenger]];
  FLTFirebaseAuthPlugin *instance = [[FLTFirebaseAuthPlugin alloc] init:registrar.messenger];

  [registrar addMethodCallDelegate:instance channel:channel];

#if TARGET_OS_OSX
  // TODO(Salakar): Publish does not exist on MacOS version of
  // FlutterPluginRegistrar.
  // TODO(Salakar): addApplicationDelegate does not exist on MacOS version of
  // FlutterPluginRegistrar. (https://github.com/flutter/flutter/issues/41471)
  FirebaseAuthHostApiSetup(registrar.messenger, instance);
  FirebaseAuthUserHostApiSetup(registrar.messenger, instance);
#else
  [registrar publish:instance];
  [registrar addApplicationDelegate:instance];
  FirebaseAuthHostApiSetup(registrar.messenger, instance);
  FirebaseAuthUserHostApiSetup(registrar.messenger, instance);
  MultiFactorUserHostApiSetup(registrar.messenger, instance);
  MultiFactoResolverHostApiSetup(registrar.messenger, instance);
#endif
}

+ (FlutterError *)convertToFlutterError:(NSError *)error {
  NSString *code = @"unknown";
  NSString *message = @"An unknown error has occurred.";

  if (error == nil) {
    return [FlutterError errorWithCode:code message:message details:@{}];
  }

  // code
  if ([error userInfo][FIRAuthErrorUserInfoNameKey] != nil) {
    // See [FIRAuthErrorCodeString] for list of codes.
    // Codes are in the format "ERROR_SOME_NAME", converting below to the format
    // required in Dart. ERROR_SOME_NAME -> SOME_NAME
    NSString *firebaseErrorCode = [error userInfo][FIRAuthErrorUserInfoNameKey];
    code = [firebaseErrorCode stringByReplacingOccurrencesOfString:@"ERROR_" withString:@""];
    // SOME_NAME -> SOME-NAME
    code = [code stringByReplacingOccurrencesOfString:@"_" withString:@"-"];
    // SOME-NAME -> some-name
    code = [code lowercaseString];
  }

  // message
  if ([error userInfo][NSLocalizedDescriptionKey] != nil) {
    message = [error userInfo][NSLocalizedDescriptionKey];
  }

  NSMutableDictionary *additionalData = [NSMutableDictionary dictionary];
  // additionalData.email
  if ([error userInfo][FIRAuthErrorUserInfoEmailKey] != nil) {
    additionalData[kArgumentEmail] = [error userInfo][FIRAuthErrorUserInfoEmailKey];
  }
  // additionalData.authCredential
  if ([error userInfo][FIRAuthErrorUserInfoUpdatedCredentialKey] != nil) {
    FIRAuthCredential *authCredential = [error userInfo][FIRAuthErrorUserInfoUpdatedCredentialKey];
    additionalData[@"authCredential"] =
        [FLTFirebaseAuthPlugin getNSDictionaryFromAuthCredential:authCredential];
  }

  // Manual message overrides to ensure messages/codes matches other platforms.
  if ([message isEqual:@"The password must be 6 characters long or more."]) {
    message = @"Password should be at least 6 characters";
  }

  return [FlutterError errorWithCode:code message:message details:additionalData];
}

+ (id)getNSDictionaryFromAuthCredential:(FIRAuthCredential *)authCredential {
  if (authCredential == nil) {
    return [NSNull null];
  }

  NSString *accessToken = nil;
  if ([authCredential isKindOfClass:[FIROAuthCredential class]]) {
    if (((FIROAuthCredential *)authCredential).accessToken != nil) {
      accessToken = ((FIROAuthCredential *)authCredential).accessToken;
    } else if (((FIROAuthCredential *)authCredential).IDToken != nil) {
      // For Sign In With Apple, the token is stored in IDToken
      accessToken = ((FIROAuthCredential *)authCredential).IDToken;
    }
  }

  return @{
    kArgumentProviderId : authCredential.provider,
    // Note: "signInMethod" does not exist on iOS SDK, so using provider
    // instead.
    kArgumentSignInMethod : authCredential.provider,
    kArgumentToken : @([authCredential hash]),
    kArgumentAccessToken : accessToken ?: [NSNull null],
  };
}

- (void)cleanupWithCompletion:(void (^)(void))completion {
  // Cleanup credentials.
  [_credentials removeAllObjects];

  for (FlutterEventChannel *channel in self->_eventChannels.allValues) {
    [channel setStreamHandler:nil];
  }
  [self->_eventChannels removeAllObjects];
  for (NSObject<FlutterStreamHandler> *handler in self->_streamHandlers.allValues) {
    [handler onCancelWithArguments:nil];
  }
  [self->_streamHandlers removeAllObjects];

  if (completion != nil) completion();
}

- (void)detachFromEngineForRegistrar:(NSObject<FlutterPluginRegistrar> *)registrar {
  [self cleanupWithCompletion:nil];
}

#pragma mark - AppDelegate

#if TARGET_OS_IPHONE
#if !__has_include(<FirebaseMessaging/FirebaseMessaging.h>)
- (BOOL)application:(UIApplication *)application
    didReceiveRemoteNotification:(NSDictionary *)notification
          fetchCompletionHandler:(void (^)(UIBackgroundFetchResult result))completionHandler {
  if ([[FIRAuth auth] canHandleNotification:notification]) {
    completionHandler(UIBackgroundFetchResultNoData);
    return YES;
  }
  return NO;
}
#endif

- (void)application:(UIApplication *)application
    didRegisterForRemoteNotificationsWithDeviceToken:(NSData *)deviceToken {
  _apnsToken = deviceToken;
}

- (BOOL)application:(UIApplication *)app openURL:(NSURL *)url options:(NSDictionary *)options {
  return [[FIRAuth auth] canHandleURL:url];
}
#endif

#pragma mark - FLTFirebasePlugin

- (void)didReinitializeFirebaseCore:(void (^_Nonnull)(void))completion {
  [self cleanupWithCompletion:completion];
}

- (NSString *_Nonnull)firebaseLibraryName {
  return LIBRARY_NAME;
}

- (NSString *_Nonnull)firebaseLibraryVersion {
  return LIBRARY_VERSION;
}

- (NSString *_Nonnull)flutterChannelName {
  return kFLTFirebaseAuthChannelName;
}

- (NSDictionary *_Nonnull)pluginConstantsForFIRApp:(FIRApp *_Nonnull)firebaseApp {
  FIRAuth *auth = [FIRAuth authWithApp:firebaseApp];
  return @{
    @"APP_LANGUAGE_CODE" : (id)[auth languageCode] ?: [NSNull null],
    @"APP_CURRENT_USER" : [auth currentUser]
        ? [[PigeonParser getPigeonDetails:[auth currentUser]] toList]
        : [NSNull null],
  };
}

#pragma mark - Firebase Auth API

// Adapted from
// https://auth0.com/docs/api-auth/tutorials/nonce#generate-a-cryptographically-random-nonce Used
// for Apple Sign In
- (NSString *)randomNonce:(NSInteger)length {
  NSAssert(length > 0, @"Expected nonce to have positive length");
  NSString *characterSet = @"0123456789ABCDEFGHIJKLMNOPQRSTUVXYZabcdefghijklmnopqrstuvwxyz-._";
  NSMutableString *result = [NSMutableString string];
  NSInteger remainingLength = length;

  while (remainingLength > 0) {
    NSMutableArray *randoms = [NSMutableArray arrayWithCapacity:16];
    for (NSInteger i = 0; i < 16; i++) {
      uint8_t random = 0;
      int errorCode = SecRandomCopyBytes(kSecRandomDefault, 1, &random);
      NSAssert(errorCode == errSecSuccess, @"Unable to generate nonce: OSStatus %i", errorCode);

      [randoms addObject:@(random)];
    }

    for (NSNumber *random in randoms) {
      if (remainingLength == 0) {
        break;
      }

      if (random.unsignedIntValue < characterSet.length) {
        unichar character = [characterSet characterAtIndex:random.unsignedIntValue];
        [result appendFormat:@"%C", character];
        remainingLength--;
      }
    }
  }

  return [result copy];
}

- (NSString *)stringBySha256HashingString:(NSString *)input {
  const char *string = [input UTF8String];
  unsigned char result[CC_SHA256_DIGEST_LENGTH];
  CC_SHA256(string, (CC_LONG)strlen(string), result);

  NSMutableString *hashed = [NSMutableString stringWithCapacity:CC_SHA256_DIGEST_LENGTH * 2];
  for (NSInteger i = 0; i < CC_SHA256_DIGEST_LENGTH; i++) {
    [hashed appendFormat:@"%02x", result[i]];
  }
  return hashed;
}

static void handleSignInWithApple(FLTFirebaseAuthPlugin *object, FIRAuthDataResult *authResult,
                                  NSString *authorizationCode, NSError *error) {
  if (error != nil) {
    if (error.code == FIRAuthErrorCodeSecondFactorRequired) {
      [object handleMultiFactorError:object.appleArguments
                          completion:object.appleCompletion
                           withError:error];
    } else {
      object.appleCompletion(nil, [FLTFirebaseAuthPlugin convertToFlutterError:error]);
    }
    return;
  }
  object.appleCompletion([PigeonParser getPigeonUserCredentialFromAuthResult:authResult
                                                           authorizationCode:authorizationCode],
                         nil);
}

- (void)authorizationController:(ASAuthorizationController *)controller
    didCompleteWithAuthorization:(ASAuthorization *)authorization
    API_AVAILABLE(macos(10.15), ios(13.0)) {
  if ([authorization.credential isKindOfClass:[ASAuthorizationAppleIDCredential class]]) {
    ASAuthorizationAppleIDCredential *appleIDCredential = authorization.credential;
    NSString *rawNonce = self.currentNonce;
    NSAssert(rawNonce != nil,
             @"Invalid state: A login callback was received, but no login request was sent.");

    if (appleIDCredential.identityToken == nil) {
      NSLog(@"Unable to fetch identity token.");
      return;
    }

    NSString *idToken = [[NSString alloc] initWithData:appleIDCredential.identityToken
                                              encoding:NSUTF8StringEncoding];
    if (idToken == nil) {
      NSLog(@"Unable to serialize id token from data: %@", appleIDCredential.identityToken);
    }

    NSString *authorizationCode = nil;
    if (appleIDCredential.authorizationCode != nil) {
      authorizationCode = [[NSString alloc] initWithData:appleIDCredential.authorizationCode
                                                encoding:NSUTF8StringEncoding];
    }

    FIROAuthCredential *credential =
        [FIROAuthProvider appleCredentialWithIDToken:idToken
                                            rawNonce:rawNonce
                                            fullName:appleIDCredential.fullName];

    if (self.isReauthenticatingWithApple == YES) {
      self.isReauthenticatingWithApple = NO;
      [[FIRAuth.auth currentUser]
          reauthenticateWithCredential:credential
                            completion:^(FIRAuthDataResult *_Nullable authResult,
                                         NSError *_Nullable error) {
                              handleSignInWithApple(self, authResult, authorizationCode, error);
                            }];

    } else if (self.linkWithAppleUser != nil) {
      [self.linkWithAppleUser
          linkWithCredential:credential
                  completion:^(FIRAuthDataResult *authResult, NSError *error) {
                    self.linkWithAppleUser = nil;
                    handleSignInWithApple(self, authResult, authorizationCode, error);
                  }];

    } else {
      FIRAuth *signInAuth =
          self.signInWithAppleAuth != nil ? self.signInWithAppleAuth : FIRAuth.auth;
      [signInAuth signInWithCredential:credential
                            completion:^(FIRAuthDataResult *_Nullable authResult,
                                         NSError *_Nullable error) {
                              self.signInWithAppleAuth = nil;
                              handleSignInWithApple(self, authResult, authorizationCode, error);
                            }];
    }
  }
}

- (void)authorizationController:(ASAuthorizationController *)controller
           didCompleteWithError:(NSError *)error API_AVAILABLE(macos(10.15), ios(13.0)) {
  NSLog(@"Sign in with Apple errored: %@", error);
  switch (error.code) {
    case ASAuthorizationErrorCanceled:
      self.appleCompletion(
          nil, [FlutterError errorWithCode:@"canceled"
                                   message:@"The user canceled the authorization attempt."
                                   details:nil]);
      break;

    case ASAuthorizationErrorInvalidResponse:
      self.appleCompletion(
          nil,
          [FlutterError errorWithCode:@"invalid-response"
                              message:@"The authorization request received an invalid response."
                              details:nil]);
      break;

    case ASAuthorizationErrorNotHandled:
      self.appleCompletion(nil,
                           [FlutterError errorWithCode:@"not-handled"
                                               message:@"The authorization request wasn’t handled."
                                               details:nil]);
      break;

    case ASAuthorizationErrorFailed:
      self.appleCompletion(nil, [FlutterError errorWithCode:@"failed"
                                                    message:@"The authorization attempt failed."
                                                    details:nil]);
      break;

    case ASAuthorizationErrorUnknown:
    default:
      self.appleCompletion(nil, [FLTFirebaseAuthPlugin convertToFlutterError:error]);
      break;
  }
  self.appleCompletion = nil;
}

- (void)handleInternalError:(nonnull void (^)(PigeonUserCredential *_Nullable,
                                              FlutterError *_Nullable))completion
                  withError:(NSError *)error {
  const NSError *underlyingError = error.userInfo[@"NSUnderlyingError"];
  if (underlyingError != nil) {
    const NSDictionary *details =
        underlyingError.userInfo[@"FIRAuthErrorUserInfoDeserializedResponseKey"];
    completion(nil, [FlutterError errorWithCode:@"internal-error"
                                        message:error.description
                                        details:details]);
    return;
  }
  completion(nil, [FlutterError errorWithCode:@"internal-error"
                                      message:error.description
                                      details:nil]);
}

- (void)handleMultiFactorError:(PigeonFirebaseApp *)app
                    completion:(nonnull void (^)(PigeonUserCredential *_Nullable,
                                                 FlutterError *_Nullable))completion
                     withError:(NSError *_Nullable)error {
#if TARGET_OS_OSX
  completion(nil, [FlutterError errorWithCode:@"second-factor-required"
                                      message:error.description
                                      details:nil]);
#else
  FIRMultiFactorResolver *resolver =
      (FIRMultiFactorResolver *)error.userInfo[FIRAuthErrorUserInfoMultiFactorResolverKey];

  NSArray<FIRMultiFactorInfo *> *hints = resolver.hints;
  FIRMultiFactorSession *session = resolver.session;

  NSString *sessionId = [[NSUUID UUID] UUIDString];
  self->_multiFactorSessionMap[sessionId] = session;

  NSString *resolverId = [[NSUUID UUID] UUIDString];
  self->_multiFactorResolverMap[resolverId] = resolver;

  NSMutableArray<NSDictionary *> *pigeonHints = [NSMutableArray array];

  for (FIRMultiFactorInfo *multiFactorInfo in hints) {
    NSString *phoneNumber;
    if ([multiFactorInfo class] == [FIRPhoneMultiFactorInfo class]) {
      FIRPhoneMultiFactorInfo *phoneFactorInfo = (FIRPhoneMultiFactorInfo *)multiFactorInfo;
      phoneNumber = phoneFactorInfo.phoneNumber;
    }

    PigeonMultiFactorInfo *object = [PigeonMultiFactorInfo
        makeWithDisplayName:multiFactorInfo.displayName
        enrollmentTimestamp:[NSNumber numberWithDouble:multiFactorInfo.enrollmentDate
                                                           .timeIntervalSince1970]
                   factorId:multiFactorInfo.factorID
                        uid:multiFactorInfo.UID
                phoneNumber:phoneNumber];

    [pigeonHints addObject:object.toList];
  }

  NSDictionary *output = @{
    kAppName : app.appName,
    kArgumentMultiFactorHints : pigeonHints,
    kArgumentMultiFactorSessionId : sessionId,
    kArgumentMultiFactorResolverId : resolverId,
  };
  completion(nil, [FlutterError errorWithCode:@"second-factor-required"
                                      message:error.description
                                      details:output]);
#endif
}

static void launchAppleSignInRequest(FLTFirebaseAuthPlugin *object, PigeonFirebaseApp *app,
                                     PigeonSignInProvider *signInProvider,
                                     void (^_Nonnull completion)(PigeonUserCredential *_Nullable,
                                                                 FlutterError *_Nullable)) {
  if (@available(iOS 13.0, macOS 10.15, *)) {
    NSString *nonce = [object randomNonce:32];
    object.currentNonce = nonce;
    object.appleCompletion = completion;
    object.appleArguments = app;

    ASAuthorizationAppleIDProvider *appleIDProvider = [[ASAuthorizationAppleIDProvider alloc] init];

    ASAuthorizationAppleIDRequest *request = [appleIDProvider createRequest];
    NSMutableArray *requestedScopes = [NSMutableArray arrayWithCapacity:2];
    if ([signInProvider.scopes containsObject:@"name"]) {
      [requestedScopes addObject:ASAuthorizationScopeFullName];
    }
    if ([signInProvider.scopes containsObject:@"email"]) {
      [requestedScopes addObject:ASAuthorizationScopeEmail];
    }
    request.requestedScopes = [requestedScopes copy];
    request.nonce = [object stringBySha256HashingString:nonce];

    ASAuthorizationController *authorizationController =
        [[ASAuthorizationController alloc] initWithAuthorizationRequests:@[ request ]];
    authorizationController.delegate = object;
    authorizationController.presentationContextProvider = object;
    [authorizationController performRequests];
  } else {
    NSLog(@"Sign in with Apple was introduced in iOS 13, update your Podfile with platform :ios, "
          @"'13.0'");
  }
}

static void handleAppleAuthResult(FLTFirebaseAuthPlugin *object, PigeonFirebaseApp *app,
                                  FIRAuth *auth, FIRAuthCredential *credentials, NSError *error,
                                  void (^_Nonnull completion)(PigeonUserCredential *_Nullable,
                                                              FlutterError *_Nullable)) {
  if (error) {
    if (error.code == FIRAuthErrorCodeSecondFactorRequired) {
      [object handleMultiFactorError:app completion:completion withError:error];
    } else {
      completion(nil, [FlutterError errorWithCode:@"sign-in-failed"
                                          message:error.localizedDescription
                                          details:error.userInfo]);
    }
    return;
  }
  if (credentials) {
    [auth signInWithCredential:credentials
                    completion:^(FIRAuthDataResult *authResult, NSError *error) {
                      if (error != nil) {
                        NSDictionary *userInfo = [error userInfo];
                        NSError *underlyingError = [userInfo objectForKey:NSUnderlyingErrorKey];

                        NSDictionary *firebaseDictionary =
                            underlyingError.userInfo[@"FIRAuthErrorUserInfoDes"
                                                     @"erializedResponseKey"];

                        if (firebaseDictionary != nil && firebaseDictionary[@"message"] != nil) {
                          // error from firebase-ios-sdk is
                          // buried in underlying error.
                          completion(nil,
                                     [FlutterError errorWithCode:@"sign-in-failed"
                                                         message:error.localizedDescription
                                                         details:firebaseDictionary[@"message"]]);
                        } else {
                          completion(nil, [FlutterError errorWithCode:@"sign-in-failed"
                                                              message:error.localizedDescription
                                                              details:error.userInfo]);
                        }
                      } else {
                        completion([PigeonParser getPigeonUserCredentialFromAuthResult:authResult
                                                                     authorizationCode:nil],
                                   nil);
                      }
                    }];
  }
}

#pragma mark - Utilities

- (void)storeAuthCredentialIfPresent:(NSError *)error {
  if ([error userInfo][FIRAuthErrorUserInfoUpdatedCredentialKey] != nil) {
    FIRAuthCredential *authCredential = [error userInfo][FIRAuthErrorUserInfoUpdatedCredentialKey];
    // We temporarily store the non-serializable credential so the
    // Dart API can consume these at a later time.
    NSNumber *authCredentialHash = @([authCredential hash]);
    _credentials[authCredentialHash] = authCredential;
  }
}

- (FIRAuth *_Nullable)getFIRAuthFromAppNameFromPigeon:(PigeonFirebaseApp *)pigeonApp {
  FIRApp *app = [FLTFirebasePlugin firebaseAppNamed:pigeonApp.appName];
  FIRAuth *auth = [FIRAuth authWithApp:app];

  auth.tenantID = pigeonApp.tenantId;

  return auth;
}

- (FIRAuthCredential *_Nullable)getFIRAuthCredentialFromArguments:(NSDictionary *)arguments
                                                              app:(PigeonFirebaseApp *)app {
  // If the credential dictionary contains a token, it means a native one has
  // been stored for later usage, so we'll attempt to retrieve it here.
  if (arguments[kArgumentToken] != nil && ![arguments[kArgumentToken] isEqual:[NSNull null]]) {
    NSNumber *credentialHashCode = arguments[kArgumentToken];
    return _credentials[credentialHashCode];
  }

  NSString *signInMethod = arguments[kArgumentSignInMethod];
  NSString *secret = arguments[kArgumentSecret] == [NSNull null] ? nil : arguments[kArgumentSecret];
  NSString *idToken =
      arguments[kArgumentIdToken] == [NSNull null] ? nil : arguments[kArgumentIdToken];
  NSString *accessToken =
      arguments[kArgumentAccessToken] == [NSNull null] ? nil : arguments[kArgumentAccessToken];
  NSString *rawNonce =
      arguments[kArgumentRawNonce] == [NSNull null] ? nil : arguments[kArgumentRawNonce];

  // Password Auth
  if ([signInMethod isEqualToString:kSignInMethodPassword]) {
    NSString *email = arguments[kArgumentEmail];
    return [FIREmailAuthProvider credentialWithEmail:email password:secret];
  }

  // Email Link Auth
  if ([signInMethod isEqualToString:kSignInMethodEmailLink]) {
    NSString *email = arguments[kArgumentEmail];
    NSString *emailLink = arguments[kArgumentEmailLink];
    return [FIREmailAuthProvider credentialWithEmail:email link:emailLink];
  }

  // Facebook Auth
  if ([signInMethod isEqualToString:kSignInMethodFacebook]) {
    return [FIRFacebookAuthProvider credentialWithAccessToken:accessToken];
  }

  // Google Auth
  if ([signInMethod isEqualToString:kSignInMethodGoogle]) {
    return [FIRGoogleAuthProvider credentialWithIDToken:idToken accessToken:accessToken];
  }

  // Twitter Auth
  if ([signInMethod isEqualToString:kSignInMethodTwitter]) {
    return [FIRTwitterAuthProvider credentialWithToken:accessToken secret:secret];
  }

  // GitHub Auth
  if ([signInMethod isEqualToString:kSignInMethodGithub]) {
    return [FIRGitHubAuthProvider credentialWithToken:accessToken];
  }

  // Phone Auth - Only supported on iOS
  if ([signInMethod isEqualToString:kSignInMethodPhone]) {
#if TARGET_OS_IPHONE
    NSString *verificationId = arguments[kArgumentVerificationId];
    NSString *smsCode = arguments[kArgumentSmsCode];
    return [[FIRPhoneAuthProvider providerWithAuth:[self getFIRAuthFromAppNameFromPigeon:app]]
        credentialWithVerificationID:verificationId
                    verificationCode:smsCode];
#else
    NSLog(@"The Firebase Phone Authentication provider is not supported on the "
          @"MacOS platform.");
    return nil;
#endif
  }

  // OAuth
  if ([signInMethod isEqualToString:kSignInMethodOAuth]) {
    NSString *providerId = arguments[kArgumentProviderId];
    return [FIROAuthProvider credentialWithProviderID:providerId
                                              IDToken:idToken
                                             rawNonce:rawNonce
                                          accessToken:accessToken];
  }

  NSLog(@"Support for an auth provider with identifier '%@' is not implemented.", signInMethod);
  return nil;
}

- (void)ensureAPNSTokenSetting {
#if !TARGET_OS_OSX
  FIRApp *defaultApp = [FIRApp defaultApp];
  if (defaultApp) {
    if ([FIRAuth auth].APNSToken == nil && _apnsToken != nil) {
      [[FIRAuth auth] setAPNSToken:_apnsToken type:FIRAuthAPNSTokenTypeUnknown];
      _apnsToken = nil;
    }
  }
#endif
}

#if !TARGET_OS_OSX
- (FIRMultiFactor *)getAppMultiFactorFromPigeon:(nonnull PigeonFirebaseApp *)app {
  FIRAuth *auth = [self getFIRAuthFromAppNameFromPigeon:app];
  FIRUser *currentUser = auth.currentUser;
  return currentUser.multiFactor;
}
#endif

- (nonnull ASPresentationAnchor)presentationAnchorForAuthorizationController:
    (nonnull ASAuthorizationController *)controller API_AVAILABLE(macos(10.15), ios(13.0)) {
#if TARGET_OS_OSX
  return [[NSApplication sharedApplication] keyWindow];
#else
  return [[UIApplication sharedApplication] keyWindow];
#endif
}

#if TARGET_OS_IPHONE

- (void)enrollPhoneApp:(nonnull PigeonFirebaseApp *)app
             assertion:(nonnull PigeonPhoneMultiFactorAssertion *)assertion
           displayName:(nullable NSString *)displayName
            completion:(nonnull void (^)(FlutterError *_Nullable))completion {
  FIRMultiFactor *multiFactor = [self getAppMultiFactorFromPigeon:app];

  FIRPhoneAuthCredential *credential =
      [[FIRPhoneAuthProvider providerWithAuth:[self getFIRAuthFromAppNameFromPigeon:app]]
          credentialWithVerificationID:[assertion verificationId]
                      verificationCode:[assertion verificationCode]];

  FIRMultiFactorAssertion *multiFactorAssertion =
      [FIRPhoneMultiFactorGenerator assertionWithCredential:credential];

  [multiFactor enrollWithAssertion:multiFactorAssertion
                       displayName:displayName
                        completion:^(NSError *_Nullable error) {
                          if (error == nil) {
                            completion(nil);
                          } else {
                            completion([FlutterError errorWithCode:@"enroll-failed"
                                                           message:error.localizedDescription
                                                           details:nil]);
                          }
                        }];
}

- (void)getEnrolledFactorsApp:(nonnull PigeonFirebaseApp *)app
                   completion:(nonnull void (^)(NSArray<PigeonMultiFactorInfo *> *_Nullable,
                                                FlutterError *_Nullable))completion {
  FIRMultiFactor *multiFactor = [self getAppMultiFactorFromPigeon:app];

  NSArray<FIRMultiFactorInfo *> *enrolledFactors = [multiFactor enrolledFactors];

  NSMutableArray<PigeonMultiFactorInfo *> *results = [NSMutableArray array];

  for (FIRMultiFactorInfo *multiFactorInfo in enrolledFactors) {
    NSString *phoneNumber;
    if ([multiFactorInfo class] == [FIRPhoneMultiFactorInfo class]) {
      FIRPhoneMultiFactorInfo *phoneFactorInfo = (FIRPhoneMultiFactorInfo *)multiFactorInfo;
      phoneNumber = phoneFactorInfo.phoneNumber;
    }

    [results
        addObject:[PigeonMultiFactorInfo
                      makeWithDisplayName:multiFactorInfo.displayName
                      enrollmentTimestamp:[NSNumber numberWithDouble:multiFactorInfo.enrollmentDate
                                                                         .timeIntervalSince1970]
                                 factorId:multiFactorInfo.factorID
                                      uid:multiFactorInfo.UID
                              phoneNumber:phoneNumber]];
  }

  completion(results, nil);
}

- (void)getSessionApp:(nonnull PigeonFirebaseApp *)app
           completion:(nonnull void (^)(PigeonMultiFactorSession *_Nullable,
                                        FlutterError *_Nullable))completion {
  FIRMultiFactor *multiFactor = [self getAppMultiFactorFromPigeon:app];
  [multiFactor getSessionWithCompletion:^(FIRMultiFactorSession *_Nullable session,
                                          NSError *_Nullable error) {
    NSString *UUID = [[NSUUID UUID] UUIDString];
    self->_multiFactorSessionMap[UUID] = session;

    PigeonMultiFactorSession *pigeonSession = [PigeonMultiFactorSession makeWithId:UUID];
    completion(pigeonSession, nil);
  }];
}

- (void)unenrollApp:(nonnull PigeonFirebaseApp *)app
          factorUid:(nonnull NSString *)factorUid
         completion:(nonnull void (^)(FlutterError *_Nullable))completion {
  FIRMultiFactor *multiFactor = [self getAppMultiFactorFromPigeon:app];
  [multiFactor unenrollWithFactorUID:factorUid
                          completion:^(NSError *_Nullable error) {
                            if (error == nil) {
                              completion(nil);
                            } else {
                              completion([FlutterError errorWithCode:@"unenroll-failed"
                                                             message:error.localizedDescription
                                                             details:nil]);
                            }
                          }];
}

- (void)resolveSignInResolverId:(nonnull NSString *)resolverId
                      assertion:(nonnull PigeonPhoneMultiFactorAssertion *)assertion
                     completion:(nonnull void (^)(PigeonUserCredential *_Nullable,
                                                  FlutterError *_Nullable))completion {
  FIRMultiFactorResolver *resolver = _multiFactorResolverMap[resolverId];

  FIRPhoneAuthCredential *credential =
      [[FIRPhoneAuthProvider provider] credentialWithVerificationID:[assertion verificationId]
                                                   verificationCode:[assertion verificationCode]];

  FIRMultiFactorAssertion *multiFactorAssertion =
      [FIRPhoneMultiFactorGenerator assertionWithCredential:credential];

  [resolver
      resolveSignInWithAssertion:multiFactorAssertion
                      completion:^(FIRAuthDataResult *_Nullable authResult,
                                   NSError *_Nullable error) {
                        if (error == nil) {
                          completion([PigeonParser getPigeonUserCredentialFromAuthResult:authResult
                                                                       authorizationCode:nil],
                                     nil);
                        } else {
                          completion(nil, [FlutterError errorWithCode:@"resolve-signin-failed"
                                                              message:error.localizedDescription
                                                              details:nil]);
                        }
                      }];
}

#endif

- (void)applyActionCodeApp:(nonnull PigeonFirebaseApp *)app
                      code:(nonnull NSString *)code
                completion:(nonnull void (^)(FlutterError *_Nullable))completion {
  FIRAuth *auth = [self getFIRAuthFromAppNameFromPigeon:app];
  [auth applyActionCode:code
             completion:^(NSError *_Nullable error) {
               if (error != nil) {
                 completion([FLTFirebaseAuthPlugin convertToFlutterError:error]);
               } else {
                 completion(nil);
               }
             }];
}

- (void)revokeTokenWithAuthorizationCodeApp:(nonnull PigeonFirebaseApp *)app
                          authorizationCode:(nonnull NSString *)authorizationCode
                                 completion:(nonnull void (^)(FlutterError *_Nullable))completion {
  FIRAuth *auth = [self getFIRAuthFromAppNameFromPigeon:app];
  [auth revokeTokenWithAuthorizationCode:authorizationCode
                              completion:^(NSError *_Nullable error) {
                                if (error != nil) {
                                  completion([FLTFirebaseAuthPlugin convertToFlutterError:error]);
                                } else {
                                  completion(nil);
                                }
                              }];
}

- (void)checkActionCodeApp:(nonnull PigeonFirebaseApp *)app
                      code:(nonnull NSString *)code
                completion:(nonnull void (^)(PigeonActionCodeInfo *_Nullable,
                                             FlutterError *_Nullable))completion {
  FIRAuth *auth = [self getFIRAuthFromAppNameFromPigeon:app];
  [auth checkActionCode:code
             completion:^(FIRActionCodeInfo *_Nullable info, NSError *_Nullable error) {
               if (error != nil) {
                 completion(nil, [FLTFirebaseAuthPlugin convertToFlutterError:error]);
               } else {
                 completion([PigeonParser parseActionCode:info], nil);
               }
             }];
}

- (void)confirmPasswordResetApp:(nonnull PigeonFirebaseApp *)app
                           code:(nonnull NSString *)code
                    newPassword:(nonnull NSString *)newPassword
                     completion:(nonnull void (^)(FlutterError *_Nullable))completion {
  FIRAuth *auth = [self getFIRAuthFromAppNameFromPigeon:app];
  [auth confirmPasswordResetWithCode:code
                         newPassword:newPassword
                          completion:^(NSError *_Nullable error) {
                            if (error != nil) {
                              completion([FLTFirebaseAuthPlugin convertToFlutterError:error]);
                            } else {
                              completion(nil);
                            }
                          }];
}

- (void)createUserWithEmailAndPasswordApp:(nonnull PigeonFirebaseApp *)app
                                    email:(nonnull NSString *)email
                                 password:(nonnull NSString *)password
                               completion:(nonnull void (^)(PigeonUserCredential *_Nullable,
                                                            FlutterError *_Nullable))completion {
  FIRAuth *auth = [self getFIRAuthFromAppNameFromPigeon:app];
  [auth createUserWithEmail:email
                   password:password
                 completion:^(FIRAuthDataResult *_Nullable authResult, NSError *_Nullable error) {
                   if (error != nil) {
                     completion(nil, [FLTFirebaseAuthPlugin convertToFlutterError:error]);
                   } else {
                     completion([PigeonParser getPigeonUserCredentialFromAuthResult:authResult
                                                                  authorizationCode:nil],
                                nil);
                   }
                 }];
}

- (void)fetchSignInMethodsForEmailApp:(nonnull PigeonFirebaseApp *)app
                                email:(nonnull NSString *)email
                           completion:(nonnull void (^)(NSArray<NSString *> *_Nullable,
                                                        FlutterError *_Nullable))completion {
  FIRAuth *auth = [self getFIRAuthFromAppNameFromPigeon:app];
  [auth fetchSignInMethodsForEmail:email
                        completion:^(NSArray<NSString *> *_Nullable providers,
                                     NSError *_Nullable error) {
                          if (error != nil) {
                            completion(nil, [FLTFirebaseAuthPlugin convertToFlutterError:error]);
                          } else {
                            completion(providers, nil);
                          }
                        }];
}

- (void)registerAuthStateListenerApp:(nonnull PigeonFirebaseApp *)app
                          completion:(nonnull void (^)(NSString *_Nullable,
                                                       FlutterError *_Nullable))completion {
  FIRAuth *auth = [self getFIRAuthFromAppNameFromPigeon:app];

  NSString *name =
      [NSString stringWithFormat:@"%@/auth-state/%@", kFLTFirebaseAuthChannelName, auth.app.name];
  FlutterEventChannel *channel = [FlutterEventChannel eventChannelWithName:name
                                                           binaryMessenger:_binaryMessenger];

  FLTAuthStateChannelStreamHandler *handler =
      [[FLTAuthStateChannelStreamHandler alloc] initWithAuth:auth];
  [channel setStreamHandler:handler];

  [_eventChannels setObject:channel forKey:name];
  [_streamHandlers setObject:handler forKey:name];

  completion(name, nil);
}

- (void)registerIdTokenListenerApp:(nonnull PigeonFirebaseApp *)app
                        completion:(nonnull void (^)(NSString *_Nullable,
                                                     FlutterError *_Nullable))completion {
  FIRAuth *auth = [self getFIRAuthFromAppNameFromPigeon:app];

  NSString *name =
      [NSString stringWithFormat:@"%@/id-token/%@", kFLTFirebaseAuthChannelName, auth.app.name];

  FlutterEventChannel *channel = [FlutterEventChannel eventChannelWithName:name
                                                           binaryMessenger:_binaryMessenger];

  FLTIdTokenChannelStreamHandler *handler =
      [[FLTIdTokenChannelStreamHandler alloc] initWithAuth:auth];
  [channel setStreamHandler:handler];

  [_eventChannels setObject:channel forKey:name];
  [_streamHandlers setObject:handler forKey:name];

  completion(name, nil);
}

- (void)sendPasswordResetEmailApp:(nonnull PigeonFirebaseApp *)app
                            email:(nonnull NSString *)email
               actionCodeSettings:(nullable PigeonActionCodeSettings *)actionCodeSettings
                       completion:(nonnull void (^)(FlutterError *_Nullable))completion {
  FIRAuth *auth = [self getFIRAuthFromAppNameFromPigeon:app];
  if (actionCodeSettings != nil) {
    FIRActionCodeSettings *settings = [PigeonParser parseActionCodeSettings:actionCodeSettings];
    [auth sendPasswordResetWithEmail:email
                  actionCodeSettings:settings
                          completion:^(NSError *_Nullable error) {
                            if (error != nil) {
                              completion([FLTFirebaseAuthPlugin convertToFlutterError:error]);
                            } else {
                              completion(nil);
                            }
                          }];
  } else {
    [auth sendPasswordResetWithEmail:email
                          completion:^(NSError *_Nullable error) {
                            if (error != nil) {
                              completion([FLTFirebaseAuthPlugin convertToFlutterError:error]);
                            } else {
                              completion(nil);
                            }
                          }];
  }
}

- (void)sendSignInLinkToEmailApp:(nonnull PigeonFirebaseApp *)app
                           email:(nonnull NSString *)email
              actionCodeSettings:(nonnull PigeonActionCodeSettings *)actionCodeSettings
                      completion:(nonnull void (^)(FlutterError *_Nullable))completion {
  FIRAuth *auth = [self getFIRAuthFromAppNameFromPigeon:app];
  [auth sendSignInLinkToEmail:email
           actionCodeSettings:[PigeonParser parseActionCodeSettings:actionCodeSettings]
                   completion:^(NSError *_Nullable error) {
                     if (error != nil) {
                       completion([FLTFirebaseAuthPlugin convertToFlutterError:error]);
                     } else {
                       completion(nil);
                     }
                   }];
}

- (void)setLanguageCodeApp:(nonnull PigeonFirebaseApp *)app
              languageCode:(nullable NSString *)languageCode
                completion:
                    (nonnull void (^)(NSString *_Nullable, FlutterError *_Nullable))completion {
  FIRAuth *auth = [self getFIRAuthFromAppNameFromPigeon:app];

  if (languageCode != nil && ![languageCode isEqual:[NSNull null]]) {
    auth.languageCode = languageCode;
  } else {
    [auth useAppLanguage];
  }

  completion(auth.languageCode, nil);
}

- (void)setSettingsApp:(nonnull PigeonFirebaseApp *)app
              settings:(nonnull PigeonFirebaseAuthSettings *)settings
            completion:(nonnull void (^)(FlutterError *_Nullable))completion {
  FIRAuth *auth = [self getFIRAuthFromAppNameFromPigeon:app];

  if (settings.userAccessGroup != nil) {
    BOOL useUserAccessGroupSuccessful;
    NSError *useUserAccessGroupErrorPtr;
    useUserAccessGroupSuccessful = [auth useUserAccessGroup:settings.userAccessGroup
                                                      error:&useUserAccessGroupErrorPtr];
    if (!useUserAccessGroupSuccessful) {
      completion([FLTFirebaseAuthPlugin convertToFlutterError:useUserAccessGroupErrorPtr]);
      return;
    }
  }

#if TARGET_OS_IPHONE
  if ([settings.appVerificationDisabledForTesting boolValue]) {
    auth.settings.appVerificationDisabledForTesting = settings.appVerificationDisabledForTesting;
  }
#else
  NSLog(@"FIRAuthSettings.appVerificationDisabledForTesting is not supported "
        @"on MacOS.");
#endif

  completion(nil);
}

- (void)signInAnonymouslyApp:(nonnull PigeonFirebaseApp *)app
                  completion:(nonnull void (^)(PigeonUserCredential *_Nullable,
                                               FlutterError *_Nullable))completion {
  FIRAuth *auth = [self getFIRAuthFromAppNameFromPigeon:app];
  [auth signInAnonymouslyWithCompletion:^(FIRAuthDataResult *authResult, NSError *error) {
    if (error != nil) {
      completion(nil, [FLTFirebaseAuthPlugin convertToFlutterError:error]);
    } else {
      completion([PigeonParser getPigeonUserCredentialFromAuthResult:authResult
                                                   authorizationCode:nil],
                 nil);
    }
  }];
}

- (void)signInWithCredentialApp:(nonnull PigeonFirebaseApp *)app
                          input:(nonnull NSDictionary<NSString *, id> *)input
                     completion:(nonnull void (^)(PigeonUserCredential *_Nullable,
                                                  FlutterError *_Nullable))completion {
  FIRAuth *auth = [self getFIRAuthFromAppNameFromPigeon:app];
  FIRAuthCredential *credential = [self getFIRAuthCredentialFromArguments:input app:app];

  if (credential == nil) {
    completion(nil, [FlutterError errorWithCode:kErrCodeInvalidCredential
                                        message:kErrMsgInvalidCredential
                                        details:nil]);
    return;
  }

  [auth signInWithCredential:credential
                  completion:^(FIRAuthDataResult *authResult, NSError *error) {
                    if (error != nil) {
                      NSDictionary *userInfo = [error userInfo];
                      NSError *underlyingError = [userInfo objectForKey:NSUnderlyingErrorKey];

                      NSDictionary *firebaseDictionary =
                          underlyingError.userInfo[@"FIRAuthErrorUserInfoDeserializedResponseKey"];

                      if (firebaseDictionary != nil && firebaseDictionary[@"message"] != nil) {
                        // error from firebase-ios-sdk is buried in underlying
                        // error.
                        completion(nil, [FlutterError errorWithCode:firebaseDictionary[@"code"]
                                                            message:firebaseDictionary[@"message"]
                                                            details:nil]);
                      } else {
                        if (error.code == FIRAuthErrorCodeSecondFactorRequired) {
                          [self handleMultiFactorError:app completion:completion withError:error];
                        } else if (error.code == FIRAuthErrorCodeInternalError) {
                          [self handleInternalError:completion withError:error];
                        } else {
                          completion(nil, [FLTFirebaseAuthPlugin convertToFlutterError:error]);
                        }
                      }
                    } else {
                      completion([PigeonParser getPigeonUserCredentialFromAuthResult:authResult
                                                                   authorizationCode:nil],
                                 nil);
                    }
                  }];
}

- (void)signInWithCustomTokenApp:(nonnull PigeonFirebaseApp *)app
                           token:(nonnull NSString *)token
                      completion:(nonnull void (^)(PigeonUserCredential *_Nullable,
                                                   FlutterError *_Nullable))completion {
  FIRAuth *auth = [self getFIRAuthFromAppNameFromPigeon:app];

  [auth signInWithCustomToken:token
                   completion:^(FIRAuthDataResult *_Nullable authResult, NSError *_Nullable error) {
                     if (error != nil) {
                       if (error.code == FIRAuthErrorCodeSecondFactorRequired) {
                         [self handleMultiFactorError:app completion:completion withError:error];
                       } else if (error.code == FIRAuthErrorCodeInternalError) {
                         [self handleInternalError:completion withError:error];
                       } else {
                         completion(nil, [FLTFirebaseAuthPlugin convertToFlutterError:error]);
                       }
                     } else {
                       completion([PigeonParser getPigeonUserCredentialFromAuthResult:authResult
                                                                    authorizationCode:nil],
                                  nil);
                     }
                   }];
}

- (void)signInWithEmailAndPasswordApp:(nonnull PigeonFirebaseApp *)app
                                email:(nonnull NSString *)email
                             password:(nonnull NSString *)password
                           completion:(nonnull void (^)(PigeonUserCredential *_Nullable,
                                                        FlutterError *_Nullable))completion {
  FIRAuth *auth = [self getFIRAuthFromAppNameFromPigeon:app];
  [auth signInWithEmail:email
               password:password
             completion:^(FIRAuthDataResult *_Nullable authResult, NSError *_Nullable error) {
               if (error != nil) {
                 if (error.code == FIRAuthErrorCodeSecondFactorRequired) {
                   [self handleMultiFactorError:app completion:completion withError:error];
                 } else if (error.code == FIRAuthErrorCodeInternalError) {
                   [self handleInternalError:completion withError:error];
                 } else {
                   completion(nil, [FLTFirebaseAuthPlugin convertToFlutterError:error]);
                 }
               } else {
                 completion([PigeonParser getPigeonUserCredentialFromAuthResult:authResult
                                                              authorizationCode:nil],
                            nil);
               }
             }];
}

- (void)signInWithEmailLinkApp:(nonnull PigeonFirebaseApp *)app
                         email:(nonnull NSString *)email
                     emailLink:(nonnull NSString *)emailLink
                    completion:(nonnull void (^)(PigeonUserCredential *_Nullable,
                                                 FlutterError *_Nullable))completion {
  FIRAuth *auth = [self getFIRAuthFromAppNameFromPigeon:app];
  [auth signInWithEmail:email
                   link:emailLink
             completion:^(FIRAuthDataResult *_Nullable authResult, NSError *_Nullable error) {
               if (error != nil) {
                 if (error.code == FIRAuthErrorCodeSecondFactorRequired) {
                   [self handleMultiFactorError:app completion:completion withError:error];
                 } else if (error.code == FIRAuthErrorCodeInternalError) {
                   [self handleInternalError:completion withError:error];
                 } else {
                   completion(nil, [FLTFirebaseAuthPlugin convertToFlutterError:error]);
                 }
               } else {
                 completion([PigeonParser getPigeonUserCredentialFromAuthResult:authResult
                                                              authorizationCode:nil],
                            nil);
               }
             }];
}

- (void)signInWithProviderApp:(nonnull PigeonFirebaseApp *)app
               signInProvider:(nonnull PigeonSignInProvider *)signInProvider
                   completion:(nonnull void (^)(PigeonUserCredential *_Nullable,
                                                FlutterError *_Nullable))completion {
  FIRAuth *auth = [self getFIRAuthFromAppNameFromPigeon:app];

  if ([signInProvider.providerId isEqualToString:kSignInMethodApple]) {
    self.signInWithAppleAuth = auth;
    launchAppleSignInRequest(self, app, signInProvider, completion);
    return;
  }
#if TARGET_OS_OSX
  NSLog(@"signInWithProvider is not supported on the "
        @"MacOS platform.");
  completion(nil, nil);
#else
  self.authProvider = [FIROAuthProvider providerWithProviderID:signInProvider.providerId auth:auth];
  NSArray *scopes = signInProvider.scopes;
  if (scopes != nil) {
    [self.authProvider setScopes:scopes];
  }
  NSDictionary *customParameters = signInProvider.customParameters;
  if (customParameters != nil) {
    [self.authProvider setCustomParameters:customParameters];
  }

  [self.authProvider
      getCredentialWithUIDelegate:nil
                       completion:^(FIRAuthCredential *_Nullable credential,
                                    NSError *_Nullable error) {
                         handleAppleAuthResult(self, app, auth, credential, error, completion);
                       }];
#endif
}

- (void)signOutApp:(nonnull PigeonFirebaseApp *)app
        completion:(nonnull void (^)(FlutterError *_Nullable))completion {
  FIRAuth *auth = [self getFIRAuthFromAppNameFromPigeon:app];

  if (auth.currentUser == nil) {
    completion(nil);
    return;
  }

  NSError *signOutErrorPtr;
  BOOL signOutSuccessful = [auth signOut:&signOutErrorPtr];

  if (!signOutSuccessful) {
    completion([FLTFirebaseAuthPlugin convertToFlutterError:signOutErrorPtr]);
  } else {
    completion(nil);
  }
}

- (void)useEmulatorApp:(nonnull PigeonFirebaseApp *)app
                  host:(nonnull NSString *)host
                  port:(nonnull NSNumber *)port
            completion:(nonnull void (^)(FlutterError *_Nullable))completion {
  FIRAuth *auth = [self getFIRAuthFromAppNameFromPigeon:app];
  [auth useEmulatorWithHost:host port:[port integerValue]];
  completion(nil);
}

- (void)verifyPasswordResetCodeApp:(nonnull PigeonFirebaseApp *)app
                              code:(nonnull NSString *)code
                        completion:(nonnull void (^)(NSString *_Nullable,
                                                     FlutterError *_Nullable))completion {
  FIRAuth *auth = [self getFIRAuthFromAppNameFromPigeon:app];

  [auth verifyPasswordResetCode:code
                     completion:^(NSString *_Nullable email, NSError *_Nullable error) {
                       if (error != nil) {
                         completion(nil, [FLTFirebaseAuthPlugin convertToFlutterError:error]);
                       } else {
                         completion(email, nil);
                       }
                     }];
}

- (void)verifyPhoneNumberApp:(nonnull PigeonFirebaseApp *)app
                     request:(nonnull PigeonVerifyPhoneNumberRequest *)request
                  completion:
                      (nonnull void (^)(NSString *_Nullable, FlutterError *_Nullable))completion {
#if TARGET_OS_OSX
  NSLog(@"The Firebase Phone Authentication provider is not supported on the "
        @"MacOS platform.");
  completion(nil, nil);
#else
  FIRAuth *auth = [self getFIRAuthFromAppNameFromPigeon:app];

  NSString *name = [NSString
      stringWithFormat:@"%@/phone/%@", kFLTFirebaseAuthChannelName, [NSUUID UUID].UUIDString];
  FlutterEventChannel *channel = [FlutterEventChannel eventChannelWithName:name
                                                           binaryMessenger:_binaryMessenger];

  NSString *multiFactorSessionId = request.multiFactorSessionId;
  FIRMultiFactorSession *multiFactorSession = nil;

  if (multiFactorSessionId != nil) {
    multiFactorSession = _multiFactorSessionMap[multiFactorSessionId];
  }

  NSString *multiFactorInfoId = request.multiFactorInfoId;

  FIRPhoneMultiFactorInfo *multiFactorInfo = nil;
  if (multiFactorInfoId != nil) {
    for (NSString *resolverId in _multiFactorResolverMap) {
      for (FIRMultiFactorInfo *info in _multiFactorResolverMap[resolverId].hints) {
        if ([info.UID isEqualToString:multiFactorInfoId] &&
            [info class] == [FIRPhoneMultiFactorInfo class]) {
          multiFactorInfo = (FIRPhoneMultiFactorInfo *)info;
          break;
        }
      }
    }
  }

#if TARGET_OS_OSX
  FLTPhoneNumberVerificationStreamHandler *handler =
      [[FLTPhoneNumberVerificationStreamHandler alloc] initWithAuth:auth];
#else
  FLTPhoneNumberVerificationStreamHandler *handler =
      [[FLTPhoneNumberVerificationStreamHandler alloc] initWithAuth:auth
                                                            request:request
                                                            session:multiFactorSession
                                                         factorInfo:multiFactorInfo];
#endif

  [channel setStreamHandler:handler];

  [_eventChannels setObject:channel forKey:name];
  [_streamHandlers setObject:handler forKey:name];

  completion(name, nil);
#endif
}

- (void)deleteApp:(nonnull PigeonFirebaseApp *)app
       completion:(nonnull void (^)(FlutterError *_Nullable))completion {
  FIRAuth *auth = [self getFIRAuthFromAppNameFromPigeon:app];
  FIRUser *currentUser = auth.currentUser;
  if (currentUser == nil) {
    completion([FlutterError errorWithCode:kErrCodeNoCurrentUser
                                   message:kErrMsgNoCurrentUser
                                   details:nil]);
    return;
  }

  [currentUser deleteWithCompletion:^(NSError *_Nullable error) {
    if (error != nil) {
      completion([FLTFirebaseAuthPlugin convertToFlutterError:error]);
    } else {
      completion(nil);
    }
  }];
}

- (void)getIdTokenApp:(nonnull PigeonFirebaseApp *)app
         forceRefresh:(nonnull NSNumber *)forceRefresh
           completion:(nonnull void (^)(PigeonIdTokenResult *_Nullable,
                                        FlutterError *_Nullable))completion {
  FIRAuth *auth = [self getFIRAuthFromAppNameFromPigeon:app];
  FIRUser *currentUser = auth.currentUser;
  if (currentUser == nil) {
    completion(nil, [FlutterError errorWithCode:kErrCodeNoCurrentUser
                                        message:kErrMsgNoCurrentUser
                                        details:nil]);
    return;
  }

  [currentUser
      getIDTokenResultForcingRefresh:[forceRefresh boolValue]
                          completion:^(FIRAuthTokenResult *tokenResult, NSError *error) {
                            if (error != nil) {
                              completion(nil, [FLTFirebaseAuthPlugin convertToFlutterError:error]);
                              return;
                            }

                            completion([PigeonParser parseIdTokenResult:tokenResult], nil);
                          }];
}

- (void)linkWithCredentialApp:(nonnull PigeonFirebaseApp *)app
                        input:(nonnull NSDictionary<NSString *, id> *)input
                   completion:(nonnull void (^)(PigeonUserCredential *_Nullable,
                                                FlutterError *_Nullable))completion {
  FIRAuth *auth = [self getFIRAuthFromAppNameFromPigeon:app];
  FIRUser *currentUser = auth.currentUser;
  if (currentUser == nil) {
    completion(nil, [FlutterError errorWithCode:kErrCodeNoCurrentUser
                                        message:kErrMsgNoCurrentUser
                                        details:nil]);
    return;
  }

  FIRAuthCredential *credential = [self getFIRAuthCredentialFromArguments:input app:app];
  if (credential == nil) {
    completion(nil, [FlutterError errorWithCode:kErrCodeInvalidCredential
                                        message:kErrMsgInvalidCredential
                                        details:nil]);
    return;
  }

  [currentUser linkWithCredential:credential
                       completion:^(FIRAuthDataResult *authResult, NSError *error) {
                         if (error != nil) {
                           if (error.code == FIRAuthErrorCodeSecondFactorRequired) {
                             [self handleMultiFactorError:app
                                               completion:completion
                                                withError:error];
                           } else {
                             completion(nil, [FLTFirebaseAuthPlugin convertToFlutterError:error]);
                           }
                         } else {
                           completion([PigeonParser getPigeonUserCredentialFromAuthResult:authResult
                                                                        authorizationCode:nil],
                                      nil);
                         }
                       }];
}

- (void)linkWithProviderApp:(nonnull PigeonFirebaseApp *)app
             signInProvider:(nonnull PigeonSignInProvider *)signInProvider
                 completion:(nonnull void (^)(PigeonUserCredential *_Nullable,
                                              FlutterError *_Nullable))completion {
  FIRAuth *auth = [self getFIRAuthFromAppNameFromPigeon:app];
  FIRUser *currentUser = auth.currentUser;
  if (currentUser == nil) {
    completion(nil, [FlutterError errorWithCode:kErrCodeNoCurrentUser
                                        message:kErrMsgNoCurrentUser
                                        details:nil]);
    return;
  }

  if ([signInProvider.providerId isEqualToString:kSignInMethodApple]) {
    self.linkWithAppleUser = currentUser;
    launchAppleSignInRequest(self, app, signInProvider, completion);
    return;
  }
#if TARGET_OS_OSX
  NSLog(@"linkWithProvider is not supported on the "
        @"MacOS platform.");
  completion(nil, nil);
#else
  self.authProvider = [FIROAuthProvider providerWithProviderID:signInProvider.providerId];
  NSArray *scopes = signInProvider.scopes;
  if (scopes != nil) {
    [self.authProvider setScopes:scopes];
  }
  NSDictionary *customParameters = signInProvider.customParameters;
  if (customParameters != nil) {
    [self.authProvider setCustomParameters:customParameters];
  }

  [currentUser
      linkWithProvider:self.authProvider
            UIDelegate:nil
            completion:^(FIRAuthDataResult *authResult, NSError *error) {
              handleAppleAuthResult(self, app, auth, authResult.credential, error, completion);
            }];
#endif
}

- (void)reauthenticateWithCredentialApp:(nonnull PigeonFirebaseApp *)app
                                  input:(nonnull NSDictionary<NSString *, id> *)input
                             completion:(nonnull void (^)(PigeonUserCredential *_Nullable,
                                                          FlutterError *_Nullable))completion {
  FIRAuth *auth = [self getFIRAuthFromAppNameFromPigeon:app];
  FIRUser *currentUser = auth.currentUser;
  if (currentUser == nil) {
    completion(nil, [FlutterError errorWithCode:kErrCodeNoCurrentUser
                                        message:kErrMsgNoCurrentUser
                                        details:nil]);
    return;
  }

  FIRAuthCredential *credential = [self getFIRAuthCredentialFromArguments:input app:app];
  if (credential == nil) {
    completion(nil, [FlutterError errorWithCode:kErrCodeInvalidCredential
                                        message:kErrMsgInvalidCredential
                                        details:nil]);
    return;
  }

  [currentUser
      reauthenticateWithCredential:credential
                        completion:^(FIRAuthDataResult *authResult, NSError *error) {
                          if (error != nil) {
                            if (error.code == FIRAuthErrorCodeSecondFactorRequired) {
                              [self handleMultiFactorError:app
                                                completion:completion
                                                 withError:error];
                            } else {
                              completion(nil, [FLTFirebaseAuthPlugin convertToFlutterError:error]);
                            }
                          } else {
                            completion(
                                [PigeonParser getPigeonUserCredentialFromAuthResult:authResult
                                                                  authorizationCode:nil],
                                nil);
                          }
                        }];
}

- (void)reauthenticateWithProviderApp:(nonnull PigeonFirebaseApp *)app
                       signInProvider:(nonnull PigeonSignInProvider *)signInProvider
                           completion:(nonnull void (^)(PigeonUserCredential *_Nullable,
                                                        FlutterError *_Nullable))completion {
  FIRAuth *auth = [self getFIRAuthFromAppNameFromPigeon:app];
  FIRUser *currentUser = auth.currentUser;
  if (currentUser == nil) {
    completion(nil, [FlutterError errorWithCode:kErrCodeNoCurrentUser
                                        message:kErrMsgNoCurrentUser
                                        details:nil]);
    return;
  }

  if ([signInProvider.providerId isEqualToString:kSignInMethodApple]) {
    self.isReauthenticatingWithApple = YES;
    launchAppleSignInRequest(self, app, signInProvider, completion);
    return;
  }
#if TARGET_OS_OSX
  NSLog(@"reauthenticateWithProvider is not supported on the "
        @"MacOS platform.");
  completion(nil, nil);
#else
  self.authProvider = [FIROAuthProvider providerWithProviderID:signInProvider.providerId];
  NSArray *scopes = signInProvider.scopes;
  if (scopes != nil) {
    [self.authProvider setScopes:scopes];
  }
  NSDictionary *customParameters = signInProvider.customParameters;
  if (customParameters != nil) {
    [self.authProvider setCustomParameters:customParameters];
  }

  [currentUser reauthenticateWithProvider:self.authProvider
                               UIDelegate:nil
                               completion:^(FIRAuthDataResult *authResult, NSError *error) {
                                 handleAppleAuthResult(self, app, auth, authResult.credential,
                                                       error, completion);
                               }];
#endif
}

- (void)reloadApp:(nonnull PigeonFirebaseApp *)app
       completion:
           (nonnull void (^)(PigeonUserDetails *_Nullable, FlutterError *_Nullable))completion {
  FIRAuth *auth = [self getFIRAuthFromAppNameFromPigeon:app];
  FIRUser *currentUser = auth.currentUser;
  if (currentUser == nil) {
    completion(nil, [FlutterError errorWithCode:kErrCodeNoCurrentUser
                                        message:kErrMsgNoCurrentUser
                                        details:nil]);
    return;
  }

  [currentUser reloadWithCompletion:^(NSError *_Nullable error) {
    if (error != nil) {
      completion(nil, [FLTFirebaseAuthPlugin convertToFlutterError:error]);
    } else {
      completion([PigeonParser getPigeonDetails:auth.currentUser], nil);
    }
  }];
}

- (void)sendEmailVerificationApp:(nonnull PigeonFirebaseApp *)app
              actionCodeSettings:(nullable PigeonActionCodeSettings *)actionCodeSettings
                      completion:(nonnull void (^)(FlutterError *_Nullable))completion {
  FIRAuth *auth = [self getFIRAuthFromAppNameFromPigeon:app];
  FIRUser *currentUser = auth.currentUser;
  if (currentUser == nil) {
    completion([FlutterError errorWithCode:kErrCodeNoCurrentUser
                                   message:kErrMsgNoCurrentUser
                                   details:nil]);
    return;
  }

  [currentUser
      sendEmailVerificationWithActionCodeSettings:[PigeonParser
                                                      parseActionCodeSettings:actionCodeSettings]

                                       completion:^(NSError *_Nullable error) {
                                         if (error != nil) {
                                           completion(
                                               [FLTFirebaseAuthPlugin convertToFlutterError:error]);
                                         } else {
                                           completion(nil);
                                         }
                                       }];
}

- (void)unlinkApp:(nonnull PigeonFirebaseApp *)app
       providerId:(nonnull NSString *)providerId
       completion:
           (nonnull void (^)(PigeonUserCredential *_Nullable, FlutterError *_Nullable))completion {
  FIRAuth *auth = [self getFIRAuthFromAppNameFromPigeon:app];
  FIRUser *currentUser = auth.currentUser;
  if (currentUser == nil) {
    completion(nil, [FlutterError errorWithCode:kErrCodeNoCurrentUser
                                        message:kErrMsgNoCurrentUser
                                        details:nil]);
    return;
  }

  [currentUser unlinkFromProvider:providerId
                       completion:^(FIRUser *_Nullable user, NSError *_Nullable error) {
                         if (error != nil) {
                           completion(nil, [FLTFirebaseAuthPlugin convertToFlutterError:error]);
                         } else {
                           completion([PigeonParser getPigeonUserCredentialFromFIRUser:user], nil);
                         }
                       }];
}

- (void)updateEmailApp:(nonnull PigeonFirebaseApp *)app
              newEmail:(nonnull NSString *)newEmail
            completion:(nonnull void (^)(PigeonUserDetails *_Nullable,
                                         FlutterError *_Nullable))completion {
  FIRAuth *auth = [self getFIRAuthFromAppNameFromPigeon:app];
  FIRUser *currentUser = auth.currentUser;
  if (currentUser == nil) {
    completion(nil, [FlutterError errorWithCode:kErrCodeNoCurrentUser
                                        message:kErrMsgNoCurrentUser
                                        details:nil]);
    return;
  }

  [currentUser updateEmail:newEmail
                completion:^(NSError *_Nullable error) {
                  if (error != nil) {
                    completion(nil, [FLTFirebaseAuthPlugin convertToFlutterError:error]);
                  } else {
                    [currentUser reloadWithCompletion:^(NSError *_Nullable reloadError) {
                      if (reloadError != nil) {
                        completion(nil, [FLTFirebaseAuthPlugin convertToFlutterError:reloadError]);
                      } else {
                        completion([PigeonParser getPigeonDetails:auth.currentUser], nil);
                      }
                    }];
                  }
                }];
}

- (void)updatePasswordApp:(nonnull PigeonFirebaseApp *)app
              newPassword:(nonnull NSString *)newPassword
               completion:(nonnull void (^)(PigeonUserDetails *_Nullable,
                                            FlutterError *_Nullable))completion {
  FIRAuth *auth = [self getFIRAuthFromAppNameFromPigeon:app];
  FIRUser *currentUser = auth.currentUser;
  if (currentUser == nil) {
    completion(nil, [FlutterError errorWithCode:kErrCodeNoCurrentUser
                                        message:kErrMsgNoCurrentUser
                                        details:nil]);
    return;
  }

  [currentUser
      updatePassword:newPassword
          completion:^(NSError *_Nullable error) {
            if (error != nil) {
              completion(nil, [FLTFirebaseAuthPlugin convertToFlutterError:error]);
            } else {
              [currentUser reloadWithCompletion:^(NSError *_Nullable reloadError) {
                if (reloadError != nil) {
                  completion(nil, [FLTFirebaseAuthPlugin convertToFlutterError:reloadError]);
                } else {
                  completion([PigeonParser getPigeonDetails:auth.currentUser], nil);
                }
              }];
            }
          }];
}

- (void)updatePhoneNumberApp:(nonnull PigeonFirebaseApp *)app
                       input:(nonnull NSDictionary<NSString *, id> *)input
                  completion:(nonnull void (^)(PigeonUserDetails *_Nullable,
                                               FlutterError *_Nullable))completion {
#if TARGET_OS_IPHONE
  FIRAuth *auth = [self getFIRAuthFromAppNameFromPigeon:app];
  FIRUser *currentUser = auth.currentUser;
  if (currentUser == nil) {
    completion(nil, [FlutterError errorWithCode:kErrCodeNoCurrentUser
                                        message:kErrMsgNoCurrentUser
                                        details:nil]);
    return;
  }

  FIRAuthCredential *credential = [self getFIRAuthCredentialFromArguments:input app:app];
  if (credential == nil) {
    completion(nil, [FlutterError errorWithCode:kErrCodeInvalidCredential
                                        message:kErrMsgInvalidCredential
                                        details:nil]);
    return;
  }

  [currentUser
      updatePhoneNumberCredential:(FIRPhoneAuthCredential *)credential
                       completion:^(NSError *_Nullable error) {
                         if (error != nil) {
                           completion(nil, [FLTFirebaseAuthPlugin convertToFlutterError:error]);
                         } else {
                           [currentUser reloadWithCompletion:^(NSError *_Nullable reloadError) {
                             if (reloadError != nil) {
                               completion(
                                   nil, [FLTFirebaseAuthPlugin convertToFlutterError:reloadError]);
                             } else {
                               completion([PigeonParser getPigeonDetails:auth.currentUser], nil);
                             }
                           }];
                         }
                       }];
#else
  NSLog(@"Updating a users phone number via Firebase Authentication is only "
        @"supported on the iOS "
        @"platform.");
  completion(nil, nil);
#endif
}

- (void)updateProfileApp:(nonnull PigeonFirebaseApp *)app
                 profile:(nonnull PigeonUserProfile *)profile
              completion:(nonnull void (^)(PigeonUserDetails *_Nullable,
                                           FlutterError *_Nullable))completion {
  FIRAuth *auth = [self getFIRAuthFromAppNameFromPigeon:app];
  FIRUser *currentUser = auth.currentUser;
  if (currentUser == nil) {
    completion(nil, [FlutterError errorWithCode:kErrCodeNoCurrentUser
                                        message:kErrMsgNoCurrentUser
                                        details:nil]);
    return;
  }

  FIRUserProfileChangeRequest *changeRequest = [currentUser profileChangeRequest];

  if ([profile.displayNameChanged boolValue]) {
    changeRequest.displayName = profile.displayName;
  }

  if ([profile.photoUrlChanged boolValue]) {
    if (profile.photoUrl == nil) {
      // We apparently cannot set photoURL to nil/NULL to remove it.
      // Instead, setting it to empty string appears to work.
      // When doing so, Dart will properly receive `null` anyway.
      changeRequest.photoURL = [NSURL URLWithString:@""];
    } else {
      changeRequest.photoURL = [NSURL URLWithString:profile.photoUrl];
    }
  }

  [changeRequest commitChangesWithCompletion:^(NSError *error) {
    if (error != nil) {
      completion(nil, [FLTFirebaseAuthPlugin convertToFlutterError:error]);
    } else {
      [currentUser reloadWithCompletion:^(NSError *_Nullable reloadError) {
        if (reloadError != nil) {
          completion(nil, [FLTFirebaseAuthPlugin convertToFlutterError:reloadError]);
        } else {
          completion([PigeonParser getPigeonDetails:auth.currentUser], nil);
        }
      }];
    }
  }];
}

- (void)verifyBeforeUpdateEmailApp:(nonnull PigeonFirebaseApp *)app
                          newEmail:(nonnull NSString *)newEmail
                actionCodeSettings:(nullable PigeonActionCodeSettings *)actionCodeSettings
                        completion:(nonnull void (^)(FlutterError *_Nullable))completion {
  FIRAuth *auth = [self getFIRAuthFromAppNameFromPigeon:app];
  FIRUser *currentUser = auth.currentUser;
  if (currentUser == nil) {
    completion([FlutterError errorWithCode:kErrCodeNoCurrentUser
                                   message:kErrMsgNoCurrentUser
                                   details:nil]);
    return;
  }

  [currentUser
      sendEmailVerificationBeforeUpdatingEmail:newEmail
                            actionCodeSettings:[PigeonParser
                                                   parseActionCodeSettings:actionCodeSettings]
                                    completion:^(NSError *error) {
                                      if (error != nil) {
                                        completion(
                                            [FLTFirebaseAuthPlugin convertToFlutterError:error]);
                                      } else {
                                        completion(nil);
                                      }
                                    }];
}

@end
